% This module encodes mechanisms for deriving knowledge about the world.
% These provide an ability to define more complex concepts and roles 
% from those in modules like startrek and simple_world, and to do subsumption
% based reasoning on such concepts and roles. The paper ``Some uses of
% higher-order logic in computational linguistics'' describes most of these
% mechanisms.

module reasoning.

import kltypes.

% Knowledge base predicates
type intrp       bool -> o.
type subsume      (i -> bool) -> (i -> bool) -> o.
type concept      (i -> bool) -> o.
type role         (i -> i -> bool) -> o.

% Some simple rules for obtaining complex concepts from simple ones.
concept C :- prim_concept C.
concept (X\ (C1 X & C2 X)) :- concept C1, concept C2.
concept (X\ (all Y\ (R X Y ==> C1 Y))) :- concept C1, role R.
concept (Z\ (some (Y\ (R Z Y)))) :- role R.
%  This is questionable since it places a role in a positive situation.
%    It seems to be required to represent the urgent-message concept
concept (U\ (all X\ (all Y\ ((R1 U X & R2 U Y) ==> R3 X Y)))) :-
  role R1, role R2, role R3.


% Some simple rules for obtaining complex roles from simple ones.
role R :- prim_role R.
role (X\Y\ (some Z\ (R1 X Z & R2 Z Y))) :- role R1, role R2.
role (X\Y\ (some A\ (C A & R1 A X & R2 A Y))) :- role R1, role R2, concept C.


%  Subsumption of concepts
subsumes C1 C2 :- prim_subsume C1 C2.
subsumes (Z\ (A Z & B Z)) C :- subsumes A C, subsumes B C.
subsumes A (Z\ (B Z & C Z)) :- subsumes A B.
subsumes A (Z\ (B Z & C Z)) :- subsumes A C.
subsumes (Z\ (all (Y\ (R Z Y ==> A Y))))
         (Z\ (all (Y\ (R Z Y ==> B Y)))) :- subsumes A B.
subsumes C C.

%  A simple little theorem prover which uses the knowledge
%  representation database.

intrp A :- fact A.
intrp (A & B)  :- intrp A, intrp B.
intrp (C X) :-   concept C, subsumes C D, intrp (D X).
intrp (P (iota Q)) :- concept P, concept Q, subsumes P Q.
intrp (C U) :-
    concept C,
    subsumes (X\ (all Y\ (R X Y ==> C Y))) D,
    intrp (R V U), intrp (D V).


select P [X|L] X :- intrp (P X).
select P [Y|L] X :- select P L X.


