%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%	CAM : Categorical Abstract Machine		  %%%
%%%	Evaluator (as defined in figure 5 of Kahn's	  %%%
%%%	STACS paper)					  %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

module cameval.

import camdecl.

type  exec	  prog -> V -> o.	   %%% execute program; result is V
type  eval	  stk -> com -> stk -> o.  %%% S1 |- C => S2
type  evalop	  oper -> val -> val -> o. %%% primitive binary ops
type  init_stack  stk -> o.		   %%% the initial stack for CAM


%%% (1) Program execution
exec (program C) V :-
	init_stack S,			%%% get the intial stack 
	eval S C S1,			%%% evaluate program
	S1 = (stk [V|Rest]).		%%% result is left on top of stack



eval (stk [(cfix P)|S]) Com S1 :- !,	%%% Unfold Recursive Environment 
	eval (stk [(P (cfix P))|S])     %%% (if its on top of stack)
	     Com S1.

eval S skip S :- !.			%%% (2) Skip

eval S (C1 & C2) S2 :- !,		%%% (3) Composition
	eval S  C1 S1,
	eval S1 C2 S2, !.

eval (stk [A|S]) (quote V) (stk [V|S]) :- !.	%%% (4) Quote

eval (stk [(pr A B)|S]) cfst (stk [A|S]) :- !.	%%% (5) Cfst

eval (stk [(pr A B)|S]) csnd (stk [B|S]) :- !.	%%% (6) Csnd

eval (stk [A|S]) push (stk [A,A|S]) :- !.	%%% (7) Push

eval (stk [B,A|S]) cons (stk [(pr A B)|S]) :- !.%%% (8) Cons

eval (stk [A,B|S]) swap (stk [B,A|S]) :- !.	%%% (9) Swap

eval (stk [A|S]) (op Op) (stk [B|S]) :- !,	%%% (10) Op
	evalop Op A B, !.

eval (stk [ctt|S]) (branch C1 C2) S1 :- !,	%%% (11) Branch (true)
	eval (stk S) C1 S1, !.

eval (stk [cff|S]) (branch C1 C2) S1 :- !,	%%% (12) Branch (false)
	eval (stk S) C2 S1, !.

eval (stk [P|S]) (cur C)			%%% (13) Cur
     (stk [(closure C P)|S]) :- !.	

eval (stk [(pr (closure C P) A)|S]) app S1 :- !,%%% (14) Apply 
	eval (stk [(pr P A)|S])  C  S1, !.

eval (stk [P|S]) (rec C) 			%%% (15) Recursion
     (stk [(cfix PP)|S]) :- !,
	pi P1\ (eval (stk [(pr P P1)|S])  C  (stk [(PP P1)|S])), !.


%%%	Builtin Operators    %%%

evalop cequal (pr A B) V :- !,
	((A = B), (V = ctt), !; (V = cff)), !.
evalop cplus  (pr (## N1) (## N2)) (## N) :- !,
	N is N1 + N2, !.
evalop cminus  (pr (## N1) (## N2)) (## N) :- !,
	N is N1 - N2, !.
evalop ctimes  (pr (## N1) (## N2)) (## N) :- !,
	N is N1 * N2, !.
evalop cequal (pr A B) V :- !,
	((A = B), (V = ctt), !; (V = cff)), !.
evalop cnequal (pr A B) V :- !,
	((A = B), (V = cff), !; (V = ctt)), !.
evalop clt  (pr (## N1) (## N2)) V :- !,
	(truth is (N1 < N2), V = ctt, !; V = cff), !.
evalop cgt  (pr (## N1) (## N2)) V :- !,
	(truth is (N1 > N2), V = ctt, !; V = cff), !.

evalop ccons (pr A B) (clist A B) :- !.
evalop chd (clist A B) A :- !.
evalop ctl (clist A B) B :- !.
evalop cempty (clist A B) cff :- !.
evalop cempty cnil ctt :- !.

%%%	The initial stack:    %%%

init_stack (stk [empty]).
