%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%        mini-ML : Evaluation Procedure                 %%%
%%%        (eval Exp V) succeeds if Exp evaluates to V.   %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

module mleval.

import mldecl lists.

type eval             tm -> tm -> o.
type evalbinaryop     tm -> tm -> tm -> tm -> o.
type evalunaryop      tm -> tm -> tm -> o.


%%% Numbers in parentheses correspond to the numbering given to the
%%% rules in figure 5.

eval (# N)  (# N).                               %%% (1) Numbers

eval tt tt.                                      %%% (2,3) constants
eval ff ff.
eval nil nil.

eval (cond E1 E2 E3) C :-                        %%% (4,5) Conditional
        eval E1 B,
        (B = tt, eval E2 C, ! ;
                 eval E3 C).

eval (E1 && E2) (C1 && C2) :-                    %%% (6) Product
        eval E1 C1,
        eval E2 C2.

eval (lamb M) (lamb M).                          %%% (9) Abstraction

eval (E1 @ E2) C :-                              %%% (10) Application
        eval E1 (lamb M),
        eval E2 C2,
        eval (M C2) C.        

eval (let M E2) C :-                             %%% (11) Let
        eval E2 C2,
        eval (M C2) C.

eval (fix M) C :-                                %%% (12) Recursion
        eval (M (fix M)) C.


%%% Primitive unary operators %%%
eval (Op @ E1) C :-
        member Op [fst, snd, hd, tl, empty],
        eval E1 C1,
        evalunaryop Op C1 C.

%%%  Primitive binary operators %%%
eval (Op @ E1 @ E2) C :-
        member Op [::, ==, !=, lt, gt, plus, minus, times],
        eval E1 C1,
        eval E2 C2,
        evalbinaryop Op C1 C2 C.

evalunaryop fst (C1 && C2) C1.                   %%% (7) First Projection
evalunaryop snd (C1 && C2) C2.                   %%% (8) Second Projection
evalunaryop hd  (:: @ C1 @ C2) C1.               %%% head of list
evalunaryop tl  (:: @ C1 @ C2) C2.               %%% tail of list
evalunaryop empty nil tt.                        %%% empty-list function
evalunaryop empty (:: @ C1 @ C2) ff.             %%%   ""
        
evalbinaryop :: C1 C2 (:: @ C1 @ C2).            %%% list construction
evalbinaryop == C C tt.                          %%% term equality
evalbinaryop == C1 C2 ff.                        %%%   ""
evalbinaryop != C C ff.                          %%% term inequality
evalbinaryop != C1 C2 tt.                        %%%   ""
evalbinaryop lt (# N1) (# N2) T :-               %%% integer <
        (truth is (N1 < N2), T = tt, ! ; T = ff).
evalbinaryop gt (# N1) (# N2) T :-               %%% integer >
        (truth is (N1 > N2), T = tt, ! ; T = ff).
evalbinaryop plus (# N1) (# N2) (# N) :-         %%% integer +
        N is (N1 + N2).
evalbinaryop minus (# N1) (# N2) (# N) :-        %%% integer -
        N is (N1 - N2).
evalbinaryop times (# N1) (# N2) (# N) :-        %%% integer *
        N is (N1 * N2).

%%%   End of module mleval   %%%

