%  File: constraints.pro

%  This file contains two methods to process flex-flex pairs.
%  
%  clean_ff_pairs takes a flex-flex pair list and returns another one
%  (possibly empty) and a subsitutions.  This operation puts the
%  flex-flex pairs in `normal form 1' and `normal form 2' (described
%  below).  The resulting flex-flex pairs are equivalent to the original
%  ones in that the set of solutions does not change.  There are two ways
%  that the lambda Prolog user can call clean_ff_pairs.  First, the
%  special goal `clean_ff' directly calls this function.  Second,
%  answering `c' after an answer subsitution is printed will redisplay
%  the answer substitution after having called clean_ff_pairs.
%  
%  triv_ff_subs takes a flex-flex pair and solves them using the trivial
%  substitution describe in Huet's TCS 1975 paper:  project the functions
%  to a new variable of primitive type.  This is not a particularly
%  interesting solution, but it can be used to remove any flex-flex pairs
%  in an answer substitution.  The only way to call triv_ff_subs is to
%  use the `t' option after the answer substitution is printed.  It is
%  possible to backtrack out of the substitution produced.
%  
%  Examples illustrating the use of these features are at the end of this
%  file.

clean_ff_pairs([],[],[]) :- !.

clean_ff_pairs(FFin,FFout,CS) :-
	hn_list(FFin,HNin), 
	nf1_list(HNin,HNtemp), 
	hn_list(NFtemp,HNtemp),
	( find_ff_subs(HNtemp,NFtemp,NFtemp1,Subs), !,
		apply_subs_to_pairlist(Subs,NFtemp1,NFtemp2),
                clean_ff_pairs(NFtemp2,FFout,CStemp),
                compose_subs(Subs,CStemp,CS) ;
	   FFout = NFtemp, CS = []
	).


% Converting a list of flex-flex pairs in hn form into its ``first 
% normal'' form described below. Pair consisting of terms that eta 
% reduce to the same variable are also removed.

nf1_list([],[]).
nf1_list([[F1,F2]|FFlist],NFFlist) :-
	nf1(F1,F2,NF1,NF2), !, 
	( eta_reduce_cv(NF1,V), eta_reduce_cv(NF2,V), !, 
			nf1_list(FFlist,NFFlist) ;
	  nf1_list(FFlist,NFFlist1), 
			NFFlist = [[NF1,NF2] | NFFlist1]
	).

% Compute the ``first normal'' form for a disagreement pair:  If 
% corresponding variables have vacuous scopes, then drop them 
% from the binders.  Input terms are assumed to be in hn form.

nf1(hn([X|B1],H1,A1),hn([Y|B2],H2,A2),F1,F2) :-
	not(free_occ_list(X,A1)), not(free_occ_list(Y,A2)), !,
	nf1(hn(B1,H1,A1),hn(B2,H2,A2),F1,F2).
nf1(hn([X|B1],H1,A1),hn([Y|B2],H2,A2),hn([X|B3],H3,A3),hn([Y|B4],H4,A4)) :- 
	!, nf1(hn(B1,H1,A1),hn(B2,H2,A2),hn(B3,H3,A3),hn(B4,H4,A4)).
nf1(hn(B1,H1,A1),hn(B2,H2,A2),hn(B1,H1,A1),hn(B2,H2,A2)).



find_ff_subs([[hn(B1,V1,A1),hn(B2,V2,A2)]|_],[[T1,T2]|R],R,[[V1,Term]]) :-
	not (V1 == V2), not (free_occ_list(V1,A2)), 
	permutation(B1,A1), !,
	new_vars_from_vars(B1,NA1),
	permute_vars(B1,A1,NA1,NB1),
	huet_normal(Term1,hn(NB1,T2,NA1)), 
	lnorm(Term1,Term), !.
find_ff_subs([[hn(B1,V1,A1),hn(B2,V2,A2)]|_],[[T1,T2]|R],R,[[V2,Term]]) :-
	not (V2 == V1), not (free_occ_list(V2,A1)), 
	permutation(B2,A2), !,
	new_vars_from_vars(B2,NA2), 
	permute_vars(B2,A2,NA2,NB2),
	huet_normal(Term1,hn(NB2,T1,NA2)), 
	lnorm(Term1,Term), !.
find_ff_subs([[H1,H2]|_],L,L,[[Var,Term]]) :-
	(nf2_l(H1,H2,Var,Term) ; nf2_l(H2,H1,Var,Term)), !.
find_ff_subs([_|HNR],[FFP|FFlist],[FFP|NFFlist],Subs) :-
	!, find_ff_subs(HNR,FFlist,NFFlist,Subs).


% Compute a second normal form substitution for a flex-flex 
% pair: If a variable in the first term's binder is vacuous 
% and the corresponding variable occurs in a rigid path in 
% one of the second term's arguments, then project away that
% argument position.

nf2_l(hn(B1,H1,A1),hn(B2,H2,A2),H2,SubsTerm) :-
	memb_and_rest_par(X,B1,B1rest,Y,B2,B2rest),
	not(free_occ_list(X,A1)),
	memb(S,A2),
	rigid_path(Y,S,B2rest), !,
	type_of(H2,Type), 
	nf2_l_subterm(S,A2,Type,SubsTerm), !.


% Select two members from two lists in parallel and return the
% remaining parts of the lists.  This can resucceed.

memb_and_rest_par(X,[X|L],L,Y,[Y|K],K).
memb_and_rest_par(X,[U|L],[U|L1],Y,[V|K],[V|K1]) :-
	memb_and_rest_par(X,L,L1,Y,K,K1).


% Building the substitution term for projecting away a chosen 
% argument.
nf2_l_subterm(Arg,Args,Type,Term) :-
	nf2_binder_and_args(Arg,Args,Binder,NArgs),
	new_variable(cv(_,Ty,v),H),
	huet_normal(Term,hn(Binder,H,NArgs)),
	type_of(Term,Type), !.	

nf2_binder_and_args(A,[A1|RArgs],[W|RBinder],RBinder) :- 
	A == A1, !,
	nv_from_term(A1,W),
	nv_from_terms(RArgs,RBinder), !.
nf2_binder_and_args(A,[A1|RArgs],[W|RBinder],[W|Ws]) :- 
	!,
	nv_from_term(A1,W), 
	nf2_binder_and_args(A,RArgs,RBinder,Ws), !.


% Generating a new variable with the same type as a given term 
% (list of terms).
nv_from_term(Term,Nv) :-
	type_of(Term,Type), new_variable(cv(_,Type,v),Nv).

nv_from_terms([],[]).
nv_from_terms([Term|L],[Nv|Nvs]) :-
	nv_from_term(Term,Nv), !, nv_from_terms(L,Nvs).


%testff :-
%      write(' ?- '), read_line(Line),
%      parse_term(main,'.',TTerm,[],[],[],_,[],_,[],_,Line,Err,Line,[]),
%      Err = false, TTerm = [F1,A],
%      write(' ?- '), read_line(Line1),
%      parse_term(main,'.',TTerm1,[],[],[],_,[],_,[],_,Line1,Err1,Line1,[]),
%      Err1 = false, TTerm1 = [F2,A],
%  clean_ff_pairs([[F1,F2]],FF_out,CS),
%  write_node(FF_out,main),
%  write_subs(CS,main).



triv_ff_subs(FFs,Subs) :-
	form_triv_subs(FFs,[],[],Subs).

form_triv_subs([],_,InSubs,InSubs).
form_triv_subs([[F1,F2]|FFs],Vlist,InSubs,OutSubs) :-
	huet_normal(F1,hn(_,V1,_)), huet_normal(F2,hn(_,V2,_)), 
	canon_subs(V1,V2,Vlist,NVlist,S1,S2),
	form_triv_subs(FFs,NVlist,[[V1,S1],[V2,S2]|Subs],OutSubs).

canon_subs(cv(_,Ty1,v),cv(_,Ty2,v),Vlist,NVlist,S1,S2) :-
	target_and_args_type(Ty1,TargTy,ArgsTy1),
	target_and_args_type(Ty2,TargTy,ArgsTy2),
	canon_var(TargTy,Vlist,NVlist,Var),
	new_vars_from_types(ArgsTy1,B1), huet_normal(S1, hn(B1,Var,[])),
	new_vars_from_types(ArgsTy2,B2), huet_normal(S2, hn(B2,Var,[])), !.

canon_var(Ty,[],[[Ty,Var]],Var) :- new_variable(cv(_,Ty,v),Var).
canon_var(Ty,[[Ty1,Var]|L],[[Ty1,Var]|L],Var) :- Ty == Ty1, !.
canon_var(Ty,[TV|L],[TV|NL],Var) :- canon_var(Ty,L,NL,Var), !.


%%%%%%%%%%%%%%%%    Examples    %%%%%%%%%%%%%%%%
%  
%  In the following example, the trivial subsitution and the cleaned
%  one are the same.
%  ?- pi X\ (pi Y\ ( (F X) = (G Y) )).
%  
%  F = F
%  G = G
%  
%   The remaining flexible - flexible pairs:
%  <Uvar2 \ Uvar1 \ (F Uvar1) , Uvar2 \ Uvar1 \ (G Uvar2)> t
%  
%  
%   The answer with the trivial solution to flex-flex pairs:
%  
%  F = Var2 \ Var1
%  G = Var3 \ Var1 ;
%  
%  
%   Backing up over the trivial solution:
%  
%  F = F
%  G = G
%  
%   The remaining flexible - flexible pairs:
%  <Uvar2 \ Uvar1 \ (F Uvar1) , Uvar2 \ Uvar1 \ (G Uvar2)> c
%  
%  
%   The answer after cleaning the flex-flex pairs:
%  F = Var6 \ Var5
%  G = Var4 \ Var5 .
%   yes


%  ?- (pi X\ (pi Y\ ( (G (X + Y)) = (F a Y)))).
%  
%  G = G
%  F = F
%  
%   The remaining flexible - flexible pairs:
%  <Uvar2 \ Uvar1 \ (G (Uvar1 + Uvar2)) , Uvar2 \ Uvar1 \ (F a Uvar2)> t
%  
%  
%   The answer with the trivial solution to flex-flex pairs:
%  
%  G = Var2 \ Var1
%  F = Var3 \ Var4 \ Var1 ;
%  
%  
%   Backing up over the trivial solution:
%  
%  G = G
%  F = F
%  
%   The remaining flexible - flexible pairs:
%  <Uvar2 \ Uvar1 \ (G (Uvar1 + Uvar2)) , Uvar2 \ Uvar1 \ (F a Uvar2)> c
%  
%  
%   The answer after cleaning the flex-flex pairs:
%  G = Var5 \ (Var9 a)
%  F = Var7 \ Var8 \ (Var9 Var7) .
%   yes

