%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                    %
%  			File:   EVAL.PRO                             %
%  		       ------------------                            %
%                                                                    %
%  This file contains the evaluator for the small functional         %
%  programming language contained in lambda Prolog.  This language   %
%  and its implementation have not been serious efforts of this      %
%  project.  The functionality which is included here is convenient  %
%  but not robust.                                                   %
%                                                                    %
%  The functional programming language, call it fp for now, has four %
%  special forms:                                                    %
%                                                                    %
%      fixpt	for recursion                                        %
%      if	the conditional		                             %
%      scons	stream constructor                                   %
%      choose	select the first answer when calling the logic       %
%  		  programming interpreter.                           %
%                                                                    %
%  The other constants that are included in this language can be     %
%  found in the module fp_sys.mod.                                   %
%                                                                    %
%  The rest of this file is a standard implementation of             %
%  call-by-value functional programming.  Values are either          %
%  "reduced" fp expressions, like integers or lists, or closures     %
%  (for lexical scoping), suspensions (for recursion and lazy        %
%  evaluation), and specials (for implementing special functions).   %
%  The latter three values can be passed around but cannot be        %
%  printed.  These should also not leave the fp interpreter for the  %
%  lp interpreter.  A future version of lambda Prolog might try to   %
%  make sense out of that possibility, however.                      %
%                                                                    %
%  It is possible to enter user defined functional expressions. See  %
%  the modules exps.mod and streams.mod in the subdirectory          %
%  examples/misc of this distribution for examples. Currently, these %
%  functions definitions are found only if they are "used" at the    %
%  top-level, that is, they are imported directly into the main      %
%  module. This use of modules is not consistent with that of the    %
%  logic programming interpreter.  Again, a future implementation    %
%  should solve this problem.		                             %
%                                                                    %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%  Fixpoint expressions

eval(Env,cv(fixpt,Ty,c) ^ (Q\A), Value) :- !,
  eval([[Q,suspend(cv(fixpt,Ty,c) ^ (Q\A),Env)]|Env],A, Value).

%  The conditional

eval(Env,cv(if,_,c) ^ P ^ T1 ^ T2, Val) :-  !,
  eval(Env,P,Truth_value),
  (Truth_value=cv(truth,bool,c),eval(Env,T1,Val),!;
                                eval(Env,T2,Val)).

%  The lazy cons operator

eval(Env,cv(scons,Ty,c)^T1^T2,cv(scons,Ty,c)^V1^suspend(T2,Env)) :- !,
  eval(Env,T1,V1).

%  The choose connection to logic programming is done here.

eval(Env,cv(choose,(Ty -> o) -> Ty,c) ^ Pred, Term) :- 
  new_variable(cv(_,Ty,v),Var),
  apply_subs_to_term(Env,Pred,P),
  lnorm(P ^ Var, Goal),
  get_imports(main,Modules),
  prove_goal(Modules,Goal,[],Subs,[],[]),!,
  apply_subs_to_term(Subs,Var,Term).
eval(Env,cv(choose,(Ty -> o) -> Ty,c) ^ Pred, cv(error,Ty,c)) :- !.

%  Consider the cases of a special function, a constant that evaluates
%  to itself, and a logical variable.

eval(_,cv(Tok,_,c), special(Tok,[],N)) :- 
  spec_fun(Tok,N), !.

eval(_,cv(Tok,Ty,c), cv(Tok,Ty,c)) :- 
  self_eval(Tok,Ty), !.

eval(Env,cv(Tok,Ty,v),Val) :- 
  (member([cv(Tok,Ty,v),Val],Env),!;
   write('Unbound logical variable:  '),
   write(Tok),write(' : '),write(Ty),nl,!,fail).

%  Now consider application and abstraction.  First application, using
%  call-by-value evaluation.

eval(Env,Var\Term,closure(Var\Term,Env)) :- !.

%  Application just calls apply.

eval(Env,T1^T2,Val) :- !,
  eval(Env,T1,Val1),
  eval(Env,T2,Val2),
  apply(Val1,Val2,Val).

%  Look up a user defined fp program in the environment.

eval(_,User_fun,Val) :-
  user_function(User_fun,Body),!,
  eval([],Body,Val).


%   Find a user defined function.

user_function(Func,Body) :-
  import(main, Mods),
  clause_instance(Mods, _, _, cv(defun,_,c), cv(defun,_,c) ^ Func ^ Body),!.

%  When running into something which is not known to have a rewriting 
%  rule, we can either return it as is or complain.  The next two 
%  clauses implement these options.  Only one can be used at a time,
%  of course.

%eval(Env,T,T) :- !.

eval(Env,T1,T2) :- !,
  write('ERROR: Unhandled eval case: '),
  nl,write(eval(Env,T1,T2)),nl,nl,fail.

apply(closure(Var\Term,Env),Val1,Val) :- !,
  eval([[Var,Val1]|Env],Term,Val).

apply(special(Const,Args,1),Val1,Val) :- !,
  do_special(Const,[Val1|Args],Val).
apply(special(Const,Args,Arity),Val,special(Const,[Val|Args],Arity1)) :- !,
  Arity1 is Arity-1.

apply(suspend(cv(fixpt,Ty,c)^(Q\A),Env),Val1,Val) :- !,
  eval([[Q,suspend(cv(fixpt,Ty,c)^(Q\A),Env)]|Env],A,Val2),
  apply(Val2,Val1,Val).

%  demand keeps forcing a suspension until it is no longer a
%  suspension.  This function is used a couple times below.
demand(suspend(Term,Env),Val) :- !, eval(Env,Term,Val1),demand(Val1,Val).
demand(V,V).

%  The rest of this file contains the declaration and implementation
%  of the various special constants in fp.

%  First declare the specials with their arity.

spec_fun(equal,		2).
spec_fun(null,		1).
spec_fun(*,		2).
spec_fun(+,		2).
spec_fun(-,		2).
spec_fun(div,		2).
spec_fun(mod,		2).
spec_fun(car,		1).
spec_fun(cdr,		1).
spec_fun(head,		1).
spec_fun(tail,		1).
spec_fun(zerop,		1).
spec_fun(<,		2).
spec_fun(=<,		2).
spec_fun(>,		2).
spec_fun(>=,		2).
spec_fun(consp,		1).
spec_fun(and,		2).
spec_fun(or,		2).
spec_fun(pair,		2).
spec_fun(first,		1).
spec_fun(second,	1).
spec_fun(pairp,		1).
spec_fun(seq,		2).
spec_fun(null_stream,	1).
spec_fun(print,		1).
spec_fun(print_sans,	1).
spec_fun(blanklines,	1).
spec_fun(Cons,		2) :- list_cons(cv(Cons,_,c)), !.

%  Declare those constants which evaltuate to themselves

self_eval(truth,	bool).
self_eval(false,	bool).
self_eval(snil,		stream ^ A).
self_eval(unit,		unit).
self_eval(N,		int)		:- integer(N).
self_eval(_,		T) 		:- T == string.
self_eval(Nil,		list ^ A) 	:- list_nil(cv(Nil,_,c)).

%
%  Boolean expressions evaluated

do_special(zerop,   [cv(0,int,c)],	cv(truth,bool,c)) :- !.
do_special(zerop,   _,			cv(false,bool,c)) :- !.
do_special(equal,   [T,T],		cv(truth,bool,c)) :- !.
do_special(equal,   _,			cv(false,bool,c)) :- !.
do_special(null,    [Nil],		cv(truth,bool,c)) :- list_nil(Nil),!.
do_special(null,    _,			cv(false,bool,c)) :- !.
do_special(consp,   [Nil],		cv(false,bool,c)) :- list_nil(Nil),!.
do_special(consp,   _,			cv(truth,bool,c)) :- !.
do_special(and,	    [cv(truth,_,c),cv(truth,_,c)], cv(truth,bool,c)).
do_special(and,	    _,			cv(false,bool,c)) :- !.
do_special(or,	    [cv(false,_,c),cv(false,_,c)], cv(false,bool,c)) :- !.
do_special(or,	    _,			cv(truth,bool,c)) :- !.

do_special(null_stream,	[S], cv(Tok,bool,c)) :- !,
  demand(S,R), ( R = cv(snil,_,c),!, Tok = truth;
                                     Tok = false ).

%
%  Integer expressions evaluated

do_special(+,	[cv(N2,_,c),cv(N1,_,c)],  cv(N3,int,c)) :-  N3 is N1+N2.
do_special(-,	[cv(N2,_,c),cv(N1,_,c)],  cv(N3,int,c)) :-  N3 is N1-N2.
do_special(*,	[cv(N2,_,c),cv(N1,_,c)],  cv(N3,int,c)) :-  N3 is N1*N2.
do_special(div, [cv(N2,_,c),cv(N1,_,c)],  cv(N3,int,c)) :-!,N3 is N1 // N2.
do_special(mod, [cv(N2,_,c),cv(N1,_,c)],  cv(N3,int,c)) :-!,N3 is N1 mod N2.
do_special(>,   [cv(N2,_,c),cv(N1,_,c)],  cv(truth,bool,c)) :- N1>N2,!.
do_special(>    ,_,			  cv(false,bool,c)) :- !.
do_special(>=,  [cv(N2,_,c),cv(N1,_,c)],  cv(truth,bool,c)) :- N1>=N2,!.
do_special(>=,  _, 			  cv(false,bool,c)) :- !.
do_special(<,   [cv(N2,_,c),cv(N1,_,c)],  cv(truth,bool,c)) :- N1<N2,!.
do_special(<,   _, 			  cv(false,bool,c)) :- !.
do_special(=<,  [cv(N2,_,c),cv(N1,_,c)],  cv(truth,bool,c)) :- N1=<N2,!.
do_special(=<,  _, 			  cv(false,bool,c)) :- !.

%
%  Stream expressions:  snil and scons were taken care of above.

do_special(head,	[L], T) :- !,
   demand(L,K),
   (K=cv(scons,_,c)^T^_,!;
    write('ERROR: head given a non-stream cell'),nl,!,fail).
do_special(tail,	[L], T) :- !,
   demand(L,K),
   (K=cv(scons,_,c)^_^T,!;
    write('ERROR: tail given a non-stream cell'),nl,!,fail).


%  Pair expressions evaluated

do_special(pair,	[T2,T1], cv(pair,Ty,c) ^ T1 ^ T2) :- !.
do_special(first,	[L], T) :- !,
   (L=cv(pair,_,c)^T^_,!;
    write('ERROR: first not given a pair'),nl,!,fail).
do_special(second,	[L], T) :- !,
   (L=cv(pair,_,c)^_^T,!;
    write('ERROR: tail not given a pair'),nl,!,fail).
do_special(pairp,	[cv(pair,_,c)^_^_], cv(truth,bool,c)) :- !.
do_special(pairp,	_, cv(false,bool,c)) :- !.

% Sequencing function

do_special(seq,	[T2,T1],cv(unit,unit,c)) :- !.

%   Output function here

do_special(print,	[T], cv(unit,unit,c)) :- !, write_term(T,main).
do_special(print_sans,	[cv(S,string,c)], cv(unit,unit,c)) :- !, write(S).
do_special(blanklines,	[cv(N,int,c)], cv(unit,unit,c)) :- !, blanklines(N).


%  List expressions evaluated

do_special(Cons,		[T2,T1], cv(Cons,Ty,c) ^ T1 ^ T2) :-
    list_cons(cv(Cons,Ty,c)), !.
do_special(car,		[L], T) :- !,
   (L=Cons^T^_, list_cons(Cons),!;
    write('ERROR: car given a non-cons cell'),nl,!,fail).
do_special(cdr,		[L], T) :- !,
   (L=Cons^_^T, list_cons(Cons),!;
    write('ERROR: cdr given a non-cons cell'),nl,!,fail).

