
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                                                         %
%                              File: FP.PRO                               %
%                             --------------                              %
%                                                                         %
%  A small functional programming interpreter is implemented in this      %
%  file and in eval.pro.  This language is usually used to evaluate       %
%  integer expressions within calls to the is predicate.  It does         %
%  contains some other niceties, such as lazy lists (streams),            %
%  higher-order programming, user definable functional programs, and an   %
%  ad hoc connection to the logic programming interpreter.                %
%                                                                         %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%  An interactive interpreter for this language can be entered by
%  calling the goal "fp" within either Prolog or lambda Prolog.

fp :- 
  setup_env(sys_defs),
  repeat, read_exp(Exp,Ty),
          fp_interp(Exp,Ty),
	  nl,
          fail.

read_exp(Exp,Ty) :-
       repeat, 
            write(' - '), read_line(Line),
            parse_term(main,'.',TTerm,[],[],[],_,[],_,[],_,Line,Err,Line,[]),
        Err = false,
        TTerm = [Exp,Ty],!.

fp_interp(cv(stop,o,c),o) :- abort.
fp_interp(cv(halt,o,c),o) :- halt.
fp_interp(cv(lp,o,c),o) :- lp.

fp_interp(Exp,Ty) :-
    eval([],Exp,Value),
    report_value(Value,Ty).

report_value(Value,Ty) :- !,
  write_term(Value,main),
  write(' : '), 
  write_type(Ty),nl.

%  This code is called by the "iss" predicate in lambda Prolog.  A
%  bag is a structure of the form 
%       X\ (X = T1 ; X = T2 ; ...; X = Tn ; fail)
%  where T1, ..., Tn are terms of the same type.  Bags are
%  higher-order versions of lists: to test membership, simply apply
%  it to an object.  This is the equivalent to bagof.

eval_bag(Mods,Ty,Pred,Var\Bag) :-
  new_variable(cv(_,Ty,v),Var),
  lnorm(Pred ^ Var, Goal),
  (bagof(Subs, prove_goal(Mods,Goal,[],Subs,[],[]),Subs_bag),!;
   Subs_bag=[]),
  make_bag(Var,Ty,Subs_bag,Bag).

make_bag(Var,Ty,[],cv(fail,o,c)).
make_bag(Var,Ty,[Subs|S],
     (cv(;,o -> o -> o,c) ^ (cv(=,Ty -> Ty -> o,c) ^ Var ^ Term) ^ Rest)) :-
  apply_subs_to_term(Subs,Var,Term),
  make_bag(Var,Ty,S,Rest).

